from .launch import launch, open_gui
from .server import *
from .server.sync import Connected

from ._patch import _patch

__all__ = ["launch", "open_gui"]

__all__ += ["get_server_address", "set_session", "get_session"]

_patch()






from functools import wraps
from flask_login import LoginManager, UserMixin, current_user, login_user, \
    logout_user, user_logged_in
from flask_session import Session
from flask_socketio import emit

__all__ = ["User", "if_user_can", "with_user_management", "listen_to_users"]

__WITH_USERS__ = False


class User(UserMixin):
    """
    Class used for users that are accessing the session through the GUI.
    """

    def __init__(self, id=None):
        self.id = id

        self.permissions = {
            "see": True,
            "edit": True,
            "get_new_plots": False,
            "manage_users": False,
            "load_session": False,
            "save_session": False
        }

    def change_permissions(self, new_permissions):
        """
        Changes the permissions of this user.

        This method can be executed only by:
        - A user that has permission to manage_users (through the GUI).
        - The launcher of the app, by executing the python method directly
        (in the console or jupyter notebook).

        Parameters
        -----------
        new_permissions: dict
            A dictionary containing new permissions that will overwrite the old ones
        """
        # If the current_user environment variable is not present, this means the
        # method is being executed from python directly
        if 'current_user' not in locals() or current_user.has_permissions("manage_users"):
            self.permissions = {**self.permissions, **new_permissions}
        else:
            raise Exception("You don't have the rights to change user permissions.")

    def has_permissions(self, *perms):
        """
        Checks if the user has the provided permissions.

        Parameters
        -----------
        *perms: str
            The permissions that you want to check. As many as you wish.

        Returns
        ---------
        boolean
            only True if it fulfills ALL the requested permissions.
        """
        for perm in perms:
            if not self.permissions[perm]:
                return False
        else:
            return True


def if_user_can(*perms):
    """
    Wrapper that restricts actions to only users with the required permissions.

    Note that the permissions of each user are stored under user.permissions.

    Parameters
    -----------
    *perms: str
        The permissions that must be fulfilled for the user to be able to perform this action.

        As many as you wish.

    Usage
    -------
    This is to be used when responding to requests using flask-socketio.

    ```
    @socketio.on("destroy the world") # Indicates which events are we listening to
    @if_user_can("use nuclear weapons") # Restricts access to only users with needed permissions
    def destroy(*args, **kwargs):
        # All the code here will only be executed if the user has permission to use nuclear weapons
    ```
    """
    def with_permissions_check(f):

        @wraps(f)
        def wrapped(*args, **kwargs):
            if __WITH_USERS__:
                if not current_user.is_authenticated:
                    emit("auth_required")
                elif current_user.has_permissions(*perms):
                    return f(*args, **kwargs)
                else:
                    raise Exception(f"You don't have the required permissions. \n These are your permissions: {current_user.permissions}")
            else:
                return f(*args, **kwargs)
        return wrapped

    return with_permissions_check


def with_user_management(app):

    # User management
    app.config['SECRET_KEY'] = 'top-secret!'

    login_manager = LoginManager(app)
    login_manager.init_app(app)

    Session(app)

    @login_manager.user_loader
    def load_user(id):
        return User(id)


def listen_to_users(socketio_on):
    """
    Sets the necessary socketio event listeners to manage users.

    Parameters
    ----------
    socketio_on: socketio.on
        The function to be used to listen to socketio events
    """
    @socketio_on('login')
    def login(username):

        login_user(User(username))

        emit('logged_in')



import logging
import sys

from sisl._environ import register_environ_variable, get_environ_variable

from . import api

SESSION = None

SERVER_HOST = None
SERVER_PORT = None

# Register the environment variables that the user can tweak
register_environ_variable("SISL_PLOTLY_API_HOST", "localhost",
                          "The host where the GUI will run when self-hosted by the user.",
                          process=str)
register_environ_variable("SISL_PLOTLY_API_PORT", 4000,
                          "The port where the GUI will run when self-hosted by the user.",
                          process=int)


def set_session(new_session, ret_old=False):
    """
    Binds a new session to the GUI.

    The old session is unbound from the GUI. If you want to keep it, store
    it in a variable (or save it to disk with `save`)before setting the new session.
    Note that, otherwise, THE OLD SESSION WILL BE LOST. You can also set "ret_old"
    to True to get a tuple of (new_session, old_session) from this method

    Parameters
    ----------
    new_session: Session
        the new session to be used by the API.
    ret_old: bool, optional
        whether the old session should also be returned.

    Returns
    ----------
    Session
        the new session, bound to the GUI. Note that if you keep using old references
        the GUI will not be in sync automatically. You need to store the returned session
        and perform all the actions on it.
    Session
        the old session, only returned if `ret_old` is `True`
    """
    global SESSION

    socketio = None

    if SESSION is not None:
        old_session = SESSION

        socketio = old_session.socketio

        old_session.socketio = None

    SESSION = new_session

    if socketio is not None:
        SESSION.socketio = socketio

    if ret_old:
        return SESSION, old_session
    return SESSION


def get_session():
    return SESSION


def get_server_address():
    return f"http://{SERVER_HOST}:{SERVER_PORT}"


def run(host=None, port=None, debug=False, app=None, socketio=None, prelaunch=None):

    global SERVER_PORT
    global SERVER_HOST

    if app is None:
        app, socketio = api.create_app(get_session, set_session)

    # Disable all kinds of logging
    if not debug:
        app.logger.disabled = True
        log = logging.getLogger("werkzeug")
        log.disabled = True
        cli = sys.modules["flask.cli"]
        cli.show_server_banner = lambda *x: None

    if prelaunch is not None:
        prelaunch(get_session, set_session)

    if host is None:
        host = get_environ_variable("SISL_PLOTLY_API_HOST")
    if port is None:
        port = get_environ_variable("SISL_PLOTLY_API_PORT")

    SERVER_HOST = host
    SERVER_PORT = port

    print(
        f"\nApi running on {get_server_address()}...\nconnect the GUI to this address or send it to someone for sharing.")

    socketio.run(app, debug=debug, host=host, port=port)




def _patch():
    from sisl.viz.plotly import Session, Plot

    from .server.sync import Connected

    # Patch the Session and Plot classes so that they work with the GUI
    # This is a bit ugly but it works for now
    def connect_class(cls):
        cls.socketio = Connected.socketio
        cls._socketio = None
        cls.autosync = Connected.autosync
        cls.emit = Connected.emit

    connect_class(Session)
    connect_class(Plot)

    # The Session class needs some extra patches to make sure the socket
    # connection is transmitted to plots
    def _transmit_socket_to_plot(session, plot, tabID=None):
        plot.socketio = session.socketio

    Session._on_plot_added = _transmit_socket_to_plot

    def _remove_socket_from_plot(session, plot):
        plot.socketio = None

    Session._on_plot_removed = _transmit_socket_to_plot

    def _transmit_socket_change(session):
        """ Transmit the socketio change to all the plots """
        for plot in session.plots.values():
            _transmit_socket_to_plot(session, plot)

    Session._on_socketio_change = _transmit_socket_change

    def _session_get_dict_for_GUI(self):
        """
        This method is thought mainly to prepare data to be sent through the API to the GUI.
        Data has to be sent as JSON, so this method can only return JSONifiable objects. (no numpy arrays, no NaN,...)
        """
        info_dict = {
            "id": self.id,
            "tabs": self.tabs,
            "settings": self.settings,
            "params": self.params,
            "paramGroups": self._param_groups,
            "updatesAvailable": self.updates_available(),
            "plotOptions": [
                {"value": subclass.__name__, "label": subclass.plot_name()}
                for subclass in self.get_plot_classes()
            ],
            "structures": self.get_structures(),
            "plotables": self.get_plotables()
        }

        return info_dict

    Session._get_dict_for_GUI = _session_get_dict_for_GUI

    # Apply also some patches to Plot
    def _plot_get_dict_for_GUI(self):
        """ This method is thought mainly to prepare data to be sent through the API to the GUI.
        Data has to be sent as JSON, so this method can only return JSONifiable objects. (no numpy arrays, no NaN,...)
        """
        from sisl.viz.plotly.input_fields import ProgramaticInput

        info_dict = {
            "id": self.id,
            "plotClass": self.__class__.__name__,
            "struct": getattr(self, "struct", None),
            "figure": self.figure,
            "settings": {param.key: self.settings[param.key] for param in self.params if not isinstance(param, ProgramaticInput)},
            "params": self.params,
            "paramGroups": self.param_groups,
            "grid_dims": getattr(self, "grid_dims", None),
            "shortcuts": self.shortcuts_for_json
        }

        return info_dict

    Plot._get_dict_for_GUI = _plot_get_dict_for_GUI


""" In this file we build the patches that will make the session
capable of updating the GUI automatically when an action is performed
on it """

from functools import wraps

from sisl._dispatcher import AbstractDispatch
from sisl.viz.plotly.plotutils import call_method_if_present

from .emiters import emit_object

__all__ = ["AutoSync", "Connected"]


class AutoSync(AbstractDispatch):
    """
    Takes care of emiting all changes automatically to the GUI.

    You probably don't need to use it directly. Inheriting from the `Connected`
    class supercharges the object by providing an autosync attribute that returns
    an instance of this class.

    AutoSync works by wrapping all methods of the object, in a way that after
    their execution, the updated session is sent to the front end of the
    graphical interface. 

    For this purpose, the wrapped object needs to have an emit method, as this is
    what AutoSync triggers after the method call. Again, the Connected class 
    automatically provides an emit method that is suitable for Session and Plot
    objects (but can be extended to others at any point).

    Note that one can avoid the behavior of AutoSync in two different ways:
        - Setting the autosync_enabled attribute to False. 
        - Passing emit=False to the executed method

    Parameters
    ----------
    obj: any
        the object that we want to autosync by using its emit method.

    Example
    ----------
    ```
    from sisl.viz import gui

    # Launch the graphical interface
    GUI.launch()

    # Now, the GUI has an associated session
    session = GUI.session
    # Since Session inherits from Connected you have an AutoSync object
    # available on its 'autosync' attribute
    session = session.autosync

    # So running any method on this session will update the GUI automatically
    session.add_tab("Tab that it's automatically seen by the GUI")

    # However, if we want to try some things before sending them, we can do
    session.autosync_enabled = False
    .
    .   Everything we do here will not be sent automatically to the GUI
    .
    session.autosync_enabled = True # And back to auto-sync

    # If you just want to avoid it for a single method, just pass emit=False
    session.add_tab("Not sent automatically", emit=False)

    ```
    """

    def __init__(self, obj):

        self.autosync_enabled = True

        super().__init__(obj)

    def dispatch(self, method):

        # We won't act if it's a private method or autosync is turned off
        if method.__name__[0] == "_" or not self.autosync_enabled:
            return method

        @wraps(method)
        def with_changes_emitted(*args, emit=True, **kwargs):

            ret = method(*args, **kwargs)

            if emit:
                self._obj.emit()

            return ret

        return with_changes_emitted

class Connected:
    """
    Helps connecting objects to the graphical interface.

    Objects that inherit from this class have the possibility
    to be synced automatically through a socketio channel.

    Parameters
    ----------
    socketio: socketio
        The socketIO channel that the object will use for transmission.

    Attributes
    ----------
    socketio: socketio
        The socketIO channel that the object will use for transmission.

    Usage
    ----------
    ```
    session.autosync.add_tab("Just a new tab")
    # This change will be automatically transmitted to all socket listeners.
    ```
    """

    def get_socketio(self):
        return self._socketio

    def set_socketio(self, new_socketio):

        self._socketio = new_socketio

        call_method_if_present(self, '_on_socketio_change')

    socketio = property(fget=get_socketio, fset=set_socketio)

    @property
    def autosync(self):
        """
        A super-powered version of the object that syncs automatically through
        the socketio channel after each method call.

        Usage
        ----------
        ```
        session.autosync.add_tab("Just a new tab")
        # This change will be automatically transmitted to all socket listeners.
        ```
        """
        return AutoSync(self)

    def emit(self, socketio=None):
        """
        Emits the object through a socketio channel.

        Parameters
        -----------
        socketio: socketio, optional
            The socketio channel through which the object must be emitted.

            NOTE: Most certainly YOU DON'T NEED TO PROVIDE A SOCKETIO, if your object
            has a socketio already associated to it, the object will be sent through 
            that channel.
        """
        if socketio is None:
            socketio = self.socketio

        emit_object(self, socketio=socketio)

        return self



from functools import partial
from pathlib import Path
from threading import Thread
import time
import webbrowser

from sisl.viz.plotly.plotutils import get_session_classes

from . import server


def open_gui():
    """
    Opens the graphical interface
    """
    webbrowser.open(str(Path(__file__).parent / "build" / "index.html"))


def launch(no_api=False, only_api=False, server_kwargs={}, load_session=None, session_cls=None, session_settings={}, interactive=False):
    """
    Launches the graphical interface.

    Parameters
    -----------
    no_api: bool, optional
        if `True`, avoids initializing the API. All it does then is to open the GUI.
    only_api: bool, optional
        whether only the api should be run. 
        You will probably want this if you already have the GUI open or are planning
        to use the online version of the GUI (https://sisl-siesta.xyz)
    server_kwargs: dict, optional
        keyword arguments that go into `sisl.viz.plotly.gui.server.run`. These are {"host", "port", "debug"}
    load_session: str or Session, optional
        the session to set.
        If it is a string, it will be interpreted as the path were the session is.

        If it is a Session, it will be used directly.

        If it is not provided, the default one will be used.
    session_cls: str or child of Session, optional
        If load_session is not provided, you can pass a class and a new session will be initialized from it.

        If it's a string it should be the name of the class.
    session_settings: dict, optional
        settings that will be applied to the new session, regardless of how it has been obtained (default, from load_session
        or from session_cls.)
    interactive: bool, optional
        whether an interactive console should be started. Probably you will never need to set this to true.
        It is only meant to open a python console in the terminal and it is used by sgui.
    """

    if no_api:
        return open_gui()

    global threads

    def prelaunch(get_session, set_session, load_session=None, session_settings={}, session_cls=None):

        session = get_session()

        if load_session is not None:
            if isinstance(load_session, str):
                load_session = load(load_session)
            set_session(load_session)
        elif session_cls is not None:
            if isinstance(session_cls, str):
                session_cls = get_session_classes().get(session_cls)
            set_session(session_cls(**session_settings))

        if session_settings:
            session.update_settings(**session_settings)

    server_kwargs = {**server_kwargs, "prelaunch": partial(prelaunch, session_settings=session_settings, load_session=load_session, session_cls=session_cls)}

    threads = [Thread(target=server.run, kwargs=server_kwargs)]

    if interactive:
        from code import interact
        #To launch an interactive console (not needed from jupyter)
        threads.append(Thread(target=interact, kwargs={'local': vars(server)}))

    if not only_api:
        threads.append(Thread(target=open_gui))

    for t in threads:
        t.start()

    print("\nThe session has started succesfully. Happy visualization!\n")

    if interactive:
        try:
            while 1:
                time.sleep(.1)
        except KeyboardInterrupt:
            print("Please use Ctrl+D to kill the interactive console first")

if __name__ == "__main__":
    run()



from flask_socketio import SocketIO, emit as real_emit
from functools import wraps

__all__ = ["emit", "emit_session", "emit_plot", "emit_loading_plot",
                "emit_loading_plot", "emit_error", "emit_object"]


def emit(*args, socketio=None, **kwargs):

    if socketio is not None:
        socketio.emit(*args, **kwargs)
    else:
        real_emit(*args, **kwargs)


def emit_session(session_to_emit=None, broadcast=True, **kwargs):
    """
    Emits a session through the socket connection

    Parameters
    -----------
    session_to_emit: Session, optional
        The session you want to emit. If none is provided, it will try to find the global session
    """
    if session_to_emit is None:
        session_to_emit = session

    return emit("current_session", session_to_emit._get_dict_for_GUI(), broadcast=broadcast, **kwargs)


def emit_plot(plot, session=None, broadcast=True, **kwargs):
    """
    Emits a plot through the socket connection

    Parameters
    -----------
    plot: str or Plot
            The plot that you want to send. 

            It can be either its ID, to search for the plot in the current session, or
            an actual plot instance.
    """
    if isinstance(plot, str):
        plot = session.plot(plot)

    emit("plot", plot._get_dict_for_GUI(), broadcast=broadcast, **kwargs)


def emit_loading_plot(plot, broadcast=True, **kwargs):
    """
    Emits a message to inform that an action is being performed on this plot.

    This is useful to make the client know that their request is on its way to be fulfilled.

    Parameters
    -----------
    plot: str or Plot
            It can be either a plot ID or an actual plot instance.
    """
    return emit("loading_plot", {"plot_id": plot if isinstance(plot, str) else plot.id}, broadcast=broadcast, **kwargs)


def emit_error(err, **kwargs):
    emit("server_error", str(err), broadcast=False, **kwargs)


def emit_object(obj, *args, **kwargs):
    """
    Emits a sisl object (plot or session) to the GUI.

    Parameters
    ----------
    obj: Plot or Session
            the object that you want to emit
    *args:
            passed directly to the corresponding emiter.
    **kwargs:
            passed directly to the corresponding emiter.
    """
    from sisl.viz import Plot, Session

    if isinstance(obj, Plot):
        emiter = emit_plot
    elif isinstance(obj, Session):
        emiter = emit_session

    emiter(obj, *args, **kwargs)



"""Configuration for the python package

Notice that most options are indicated in setup.cfg.
Here we only include options that can't be included in setup.cfg
"""
from setuptools import setup
from pathlib import Path

def readme():
    readme_path = Path("README.md")
    if not readme_path.exists():
        return ""
    
    with open(readme_path, "r") as f:
        readme_content = f.read()

    return readme_content.split("Info for developers")[0]

setup(
    package_data={
        "": ["build/*", "build/static/*/*"],
    },
    long_description=readme(),
    long_description_content_type="text/markdown",
)


import argparse

from sisl.viz.plotly.plotutils import get_session_classes

from sisl.viz.plotly._user_customs import SESSION_FILE
from .launch import launch


def general_arguments(parser):

    parser.add_argument('--only-api', dest='only_api', action="store_true",
        help="Pass this flag if all you want to do is initialize the API, but not the GUI. You can do this if you plan"+
        "to access the graphical interface by some other means (e.g. https://sisl-siesta.xyz)"
    )

    parser.add_argument('--no-api', dest='no_api', action="store_true",
        help="Pass this flag if all you want to is to open the GUI. This will probably make sense if you already have an api running"+
        " or someone else is running the API for you."
    )


def sgui():
    """
    Command line interface for launching GUI related stuff
    """
    from sisl.viz import Session

    avail_session_classes = get_session_classes()

    parser = argparse.ArgumentParser(prog='sgui',
                                     description="Command line utility to launch sisl's graphical interface.")

    general_arguments(parser)

    parser.add_argument('--load', '-l', type=str, nargs="?", default=None,
                        help='The path to the session that you want to open in the GUI. If not provided, a fresh new session will be used.')

    for param in Session._get_class_params()[0]:
        if param.dtype is not None and not isinstance(param.dtype, str):
            parser.add_argument(f'--{param.key}', type=param.parse, required=False, help=getattr(param, "help", ""))

    subparsers = parser.add_subparsers(
        help="YOU DON'T NEED TO PASS A SESSION CLASS. You can provide a session file to load a saved session (see the --load flag)."+
        " However, if you want to start a new session and the default one (BlankSession) is not good for you"+
        " you can pass a session class. By doing so, you will also get access to session-specific settings. Try sgui BlankSession -h, for example." +
        " Note that you can also build your own sessions that will be automatically available here." +
        f" Sisl is looking to import plots defined in {SESSION_FILE}",
        dest="session_class"
    )

    for name, SessionClass in avail_session_classes.items():
        doc = SessionClass.__doc__ or ""
        specific_parser = subparsers.add_parser(name, help=doc.split(".")[0])
        general_arguments(specific_parser)
        for param in SessionClass._get_class_params()[0]:
            if param.dtype is not None and not isinstance(param.dtype, str):
                specific_parser.add_argument(f'--{param.key}', type=param.parse, required=False, help=getattr(param, "help", ""))

    args = parser.parse_args()

    # Note that it doesn't matter if we include invalid settings. Configurable will just ignore them
    settings = {key: val for key, val in vars(args).items() if val is not None}

    launch(interactive=True, load_session=args.load, session_settings=settings, session_cls=args.session_class, only_api=args.only_api, no_api=args.no_api)



import shutil
import pathlib
import traceback

from plotly.graph_objects import Figure
import numpy as np
import simplejson
from simplejson.encoder import JSONEncoder
from functools import partial

import flask
from flask import Flask, request, jsonify, make_response
from flask_socketio import SocketIO, join_room

from sisl.viz.plotly.plotutils import load

from .emiters import emit_plot, emit_session, emit_error, emit_loading_plot, emit
from .user_management import with_user_management, if_user_can, listen_to_users


__all__ = ["APP", "SESSION", "SOCKETIO", "set_session", "create_app"]


__DEBUG = False


class CustomJSONEncoder(JSONEncoder):

    def default(self, obj):

        if isinstance(obj, Figure):
            return obj.to_plotly_json()
        elif hasattr(obj, "to_json"):
            return obj.to_json()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.generic):
            return obj.item()
        elif isinstance(obj, pathlib.Path):
            return str(obj)

        return super().default(obj)

# We need to use simplejson because built-in json happily parses nan to NaN
# and then javascript does not understand it
simplejson.dumps = partial(simplejson.dumps, ignore_nan=True, cls=CustomJSONEncoder)

def create_app(get_session, set_session):

    from sisl.viz.plotly import BlankSession, Plot

    app = Flask("SISL GUI API")

    # No user management yet
    if False:
        with_user_management(app)

    socketio = SocketIO(app, cors_allowed_origins="*",
                        json=simplejson, manage_session=True, async_mode="threading")
                        # async_mode="threading" this option can not use websockets, therefore there is less communication performance
                        # however, it's the only way we can emit socket events from outside of the thread that is running the api
                        # Maybe instead of threading we can use socketio.start_background_task (see https://github.com/miguelgrinberg/Flask-SocketIO/issues/876)
    on = socketio.on

    if False:
        listen_to_users(on, emit_session)

    # Create a new session
    new_session = BlankSession(socketio=socketio)
    set_session(new_session)
    new_session.socketio = socketio

    @socketio.on_error()
    def send_error(err):
        emit_error(err)
        raise err

    @on("request_session")
    @if_user_can("see")
    def send_session(path = None):

        session = get_session()

        if path is not None:
            session = load(path)
            set_session(session)

        emit_session(session, broadcast = False)

    @on("apply_method_on_session")
    @if_user_can("edit")
    def apply_method(method_name, kwargs = {}, *args):

        if __DEBUG:
            print(f"Applying {method_name}. Args: {args}. Kwargs: {kwargs}")

        if kwargs is None:
            # This is because the GUI might send None
            kwargs = {}

        # Remember that if the method is not found an error will be raised
        # but it will be handled socketio.on_error (used above)
        method = getattr(get_session().autosync, method_name)

        # Since the session is bound to the app, this will automatically emit the
        # session
        returns = method(*args, **kwargs)

        if kwargs.get("get_returns", None):
            # Let's send the returned values if the user asked for it
            event_name = kwargs.get("returns_as", "call_returns")
            emit(event_name, returns, {"method_name": method_name}, broadcast=False)

    @on("get_plot")
    @if_user_can("see")
    def retrieve_plot(plotID):
        if __DEBUG:
            print(f"Asking for plot: {plotID}")

        emit_plot(plotID, get_session(), broadcast=False)

    @on("upload_file")
    @if_user_can("edit")
    def plot_uploaded_file(file_bytes, name):

        session = get_session()

        dirname = session.get_setting("file_storage_dir")
        if not dirname.exists():
            dirname.mkdir()

        file_name = dirname / name
        with open(file_name, "wb") as fh:
            fh.write(file_bytes)

        # file_name = name
        # file_contents = {name: file_bytes}

        plot = Plot(file_name)#, attrs_for_plot={"_file_contents": file_contents}, _debug=True) #
        session.autosync.add_plot(plot, session.tabs[0]["id"])

        if not session.get_setting("keep_uploaded"):
            shutil.rmtree(str(dirname))

    return app, socketio
